/*
 *  ========================================================================
 *  DIScrete event baSed Energy Consumption simulaTor 
 *    					             for Clouds and Federations (DISSECT-CF)
 *  ========================================================================
 *  
 *  This file is part of DISSECT-CF.
 *  
 *  DISSECT-CF is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *  
 *  DISSECT-CF is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 *  General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with DISSECT-CF.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  (C) Copyright 2013, Gabor Kecskemeti (gkecskem@dps.uibk.ac.at,
 *   									  kecskemeti.gabor@sztaki.mta.hu)
 */

package hu.mta.sztaki.lpds.cloud.simulator.iaas;

import hu.mta.sztaki.lpds.cloud.simulator.DeferredEvent;
import hu.mta.sztaki.lpds.cloud.simulator.iaas.VMManager.VMManagementException;
import hu.mta.sztaki.lpds.cloud.simulator.iaas.resourcemodel.ResourceConsumption;
import hu.mta.sztaki.lpds.cloud.simulator.iaas.resourcemodel.UtilizeResources;
import hu.mta.sztaki.lpds.cloud.simulator.iaas.resourcemodel.ConsumptionEventAdapter;
import hu.mta.sztaki.lpds.cloud.simulator.storage.NetworkNode;
import hu.mta.sztaki.lpds.cloud.simulator.storage.NetworkNode.NetworkException;
import hu.mta.sztaki.lpds.cloud.simulator.storage.Repository;
import hu.mta.sztaki.lpds.cloud.simulator.storage.StorageObject;
import hu.mta.sztaki.lpds.cloud.simulator.storage.VirtualAppliance;

import java.util.ArrayList;
import java.util.EnumSet;


public class VirtualMachine extends UtilizeResources {

	public static class StateChangeException extends VMManagementException {
		private static final long serialVersionUID = 2950595344006507672L;

		public StateChangeException(final String e) {
			super(e);
		}

	}

	public static interface StateChange {
		void stateChanged(State oldState, State newState);
	}

	private class EventSetup {

		public final State expectedState;

		public EventSetup(final State eState) {
			expectedState = eState;
		}

		public void changeEvents() {
			setState(expectedState);
		}
	}

	private final EventSetup sdEvent = new EventSetup(State.SHUTDOWN);
	private final EventSetup susEvent = new EventSetup(State.SUSPENDED);

	private VirtualAppliance va;
	private ResourceConstraints rc;
	private StorageObject disk = null;
	private StorageObject savedmemory = null;
	private PhysicalMachine host = null;
	private Repository vasource = null;

	public static enum State {
		INITIAL_TR,
		STARTUP,
		RUNNING,
		SUSPEND_TR,
		SUSPENDED,
		SUSPENDED_MIG,
		RESUME_TR,
		MIGRATING,
		SHUTDOWN,
		DESTROYED
	};

	public final static EnumSet<State> consumingStates = EnumSet.of(
			State.STARTUP, State.RUNNING, State.MIGRATING, State.RESUME_TR);
	public final static EnumSet<State> transferringStates = EnumSet.of(
			State.INITIAL_TR, State.SUSPEND_TR, State.RESUME_TR,
			State.MIGRATING);
	public final static EnumSet<State> suspendedStates = EnumSet.of(
			State.SUSPENDED, State.SUSPENDED_MIG);
	public final static EnumSet<State> preStartupStates = EnumSet.of(
			State.DESTROYED, State.SHUTDOWN);

	private ArrayList<StateChange> subscribers = new ArrayList<StateChange>(5);

	private State currState = State.DESTROYED;

	public static final float loadwhilenotrunning = 0.2f;

	private ArrayList<ResourceConsumption> suspendedTasks = new ArrayList<ResourceConsumption>();

	public VirtualMachine(final VirtualAppliance va,
			final ResourceConstraints rc) {
		super(rc.requiredCPUs);
		this.va = va;
		this.rc = rc;
	}

	private void setState(final State newstate) {
		for (StateChange sc : subscribers) {
			sc.stateChanged(currState, newstate);
		}
		currState = newstate;
	}

	public VirtualAppliance getVa() {
		return va;
	}

	public ResourceConstraints getRc() {
		return rc;
	}

	public State getState() {
		return currState;
	}

	public float getLoad() {
		return consumingStates.contains(currState) ? (currState == State.RUNNING ? 1
				: loadwhilenotrunning)
				: 0;
	}

	public void prepare(final PhysicalMachine pm, final Repository vasource)
			throws StateChangeException, VMManagementException,
			NetworkNode.NetworkException {
		if (currState != State.DESTROYED) {
			throw new StateChangeException("The VM is not destroyed");
		}
		initialTransfer(pm, vasource, sdEvent);
	}

	class InitialTransferEvent extends ConsumptionEventAdapter {
		final Repository target;
		final EventSetup esetup;
		final String diskid;

		public InitialTransferEvent(final Repository t, final EventSetup event,
				final String did) {
			target = t;
			esetup = event;
			diskid = did;
		}

		@Override
		public void conComplete() {
			disk = target.lookup(diskid);
			esetup.changeEvents();
		}
	}

	private void initialTransfer(final PhysicalMachine pm,
			final Repository vasource, final EventSetup es)
			throws VMManagementException, NetworkNode.NetworkException {
		final State oldState = currState;
		setState(State.INITIAL_TR);
		host = pm;
		this.vasource = vasource;
		final String diskid = Integer.toString((hashCode()));
		final long bgnwload = va.getBgNetworkLoad();
		final boolean noerror;
		final Repository pmdisk = pm.getLocalDisk();
		if (bgnwload > 0) {
			noerror = vasource.duplicateContent(va.id, diskid,
					new InitialTransferEvent(vasource, es, diskid));
		} else {
			if (vasource == null) {
				noerror = pmdisk.duplicateContent(va.id, diskid,
						new InitialTransferEvent(pmdisk, es, diskid));
			} else {
				noerror = vasource.requestContentDelivery(va.id, diskid,
						pmdisk, new InitialTransferEvent(pmdisk, es, diskid));
			}
		}
		if (!noerror) {
			setState(oldState);
			throw new VMManagementException("Initial transfer failed");
		}
	}

	final EventSetup switchonEvent = new EventSetup(State.STARTUP) {
		@Override
		public void changeEvents() {
			super.changeEvents();
			new DeferredEvent((int) va.getStartupdelay()) {
				@Override
				protected void eventAction() {
					setState(State.RUNNING);
				}
			};
			try {
				startBGLoad();
			} catch (NetworkException e) {
				e.printStackTrace();
			}
		}
	};

	public void switchOn(final PhysicalMachine pm, final Repository vasource)
			throws StateChangeException, VMManagementException,
			NetworkNode.NetworkException {
		switch (currState) {
		case DESTROYED:
			initialTransfer(pm, vasource, switchonEvent);
			break;
		case SHUTDOWN:
			switchonEvent.changeEvents();
			break;
		default:
			throw new StateChangeException(
					"The VM is not shut down or destroyed");
		}
	}

	private void resumeAfterMigration(final PhysicalMachine pm)
			throws NetworkNode.NetworkException {
		final PhysicalMachine oldHost = host;
		host = pm;
		try {
			resume();
			final Repository previousDisk = oldHost.getLocalDisk();
			previousDisk.deregisterObject(disk);
			previousDisk.deregisterObject(savedmemory);
		} catch (StateChangeException e) {
			System.err.println("IMPROPER STATE DURING MIGRATION!");
		} catch (VMManagementException e) {
			host = oldHost;
			setState(State.SUSPENDED_MIG);
		}
	}

	private void actualMigration(final PhysicalMachine pm)
			throws NetworkNode.NetworkException {
		final boolean[] cancelMigration = new boolean[1];
		cancelMigration[0] = false;
		class MigrationEvent extends ConsumptionEventAdapter {
			int eventcounter = 0;

			@Override
			public void conComplete() {
				if (cancelMigration[0]) {
					pm.getLocalDisk().deregisterObject(disk);
				} else {
					eventcounter++;
					if (eventcounter == 2) {
						try {
							resumeAfterMigration(pm);
						} catch (NetworkException e) {
							e.printStackTrace();
						}
					}
				}
			}
		}

		for (final ResourceConsumption con : suspendedTasks) {
			con.setProvider(pm);
		}

		if (va.getBgNetworkLoad() > 0) {
			resumeAfterMigration(pm);
		} else {

			final Repository from = host.getLocalDisk();
			final Repository to = pm.getLocalDisk();
			final MigrationEvent mp = new MigrationEvent();
			if (from.requestContentDelivery(disk.id, to, mp)) {
				if (from.requestContentDelivery(savedmemory.id, to, mp)) {
					return;
				}
				cancelMigration[0] = true;
			}
			setState(State.SUSPENDED_MIG);
		}
	}

	public void migrate(final PhysicalMachine pm) throws StateChangeException,
			VMManagementException, NetworkNode.NetworkException {
		if (suspendedStates.contains(currState)) {
			actualMigration(pm);
		} else {
			if (va.getBgNetworkLoad() <= 0) {
				NetworkNode.checkConnectivity(host.getLocalDisk(),
						pm.getLocalDisk());
			}
			suspend(new EventSetup(State.SUSPENDED) {
				@Override
				public void changeEvents() {
					super.changeEvents();
					try {
						actualMigration(pm);
					} catch (NetworkException e) {
					}
				}
			});
		}
	}

	public void destroy(final boolean killTasks) throws StateChangeException {
		if (transferringStates.contains(currState)) {
			throw new StateChangeException(
					"Parts of the VM are under transfer."
							+ "This transfer should be finished before destruction.");
		}
		if (currState.equals(State.RUNNING)) {
			switchoff(killTasks);
		}
		if (host != null) {
			host.getLocalDisk().deregisterObject(disk);
			host.getLocalDisk().deregisterObject(savedmemory);
			host = null;
		}
		if (vasource != null) {
			vasource.deregisterObject(disk);
			vasource.deregisterObject(savedmemory);
			vasource = null;
		}
		setState(State.DESTROYED);
	}

	public void switchoff(final boolean killTasks) throws StateChangeException {
		if (currState != State.RUNNING) {
			throw new StateChangeException(
					"Cannot switch off a not running machine");
		}
		if (killTasks) {
			for (final ResourceConsumption con : toProcess) {
				con.cancel();
			}
		} else if (!toProcess.isEmpty()) {
			throw new StateChangeException(
					"Cannot switch off a running machine with running tasks");
		}
		setState(State.SHUTDOWN);
	}

	public void suspend() throws StateChangeException, VMManagementException,
			NetworkNode.NetworkException {
		suspend(susEvent);
	}

	private void suspend(final EventSetup ev) throws StateChangeException,
			VMManagementException, NetworkNode.NetworkException {
		if (currState != State.RUNNING) {
			throw new StateChangeException(
					"Cannot suspend a not running machine");
		}
		for (final ResourceConsumption con : toProcess) {
			con.suspend();
			suspendedTasks.add(con);
		}
		final String memid = "VM-Memory-State" + hashCode();
		final String tmemid = "Temp-" + memid;
		final Repository pmdisk = host.getLocalDisk();
		savedmemory = new StorageObject(tmemid, rc.requiredMemory, false);
		if (pmdisk.registerObject(savedmemory)) {
			throw new VMManagementException(
					"Not enough space on localDisk for the suspend operation of "
							+ hashCode());
		}
		setState(State.SUSPEND_TR);
		final long bgnwload = va.getBgNetworkLoad();
		class SuspendComplete extends ConsumptionEventAdapter {
			@Override
			public void conComplete() {
				pmdisk.deregisterObject(savedmemory);
				savedmemory = bgnwload > 0 ? vasource.lookup(memid) : pmdisk
						.lookup(memid);
				setState(ev.expectedState);
				ev.changeEvents();
			}
		}
		;
		boolean trsuccess;
		if (bgnwload > 0) {
			trsuccess = pmdisk.requestContentDelivery(tmemid, memid, vasource,
					new SuspendComplete());
		} else {
			trsuccess = pmdisk.duplicateContent(tmemid, memid,
					new SuspendComplete());
		}
		if (!trsuccess) {
			setState(State.RUNNING);
			throw new VMManagementException(
					"Not enough space on localDisk for the suspend operation of "
							+ hashCode());
		}
	}

	public void resume() throws StateChangeException, VMManagementException,
			NetworkNode.NetworkException {
		if (!suspendedStates.contains(currState)) {
			throw new StateChangeException(
					"Cannot resume a not suspended machine");
		}
		setState(State.RESUME_TR);
		final String tmemid = "Temp-" + savedmemory.id;
		final Repository pmdisk = host.getLocalDisk();
		final long bgnwload = va.getBgNetworkLoad();
		class ResumeComplete extends ConsumptionEventAdapter {
			@Override
			public void conComplete() {
				pmdisk.deregisterObject(tmemid);
				if (bgnwload > 0) {
					vasource.deregisterObject(savedmemory);
				} else {
					pmdisk.deregisterObject(savedmemory);
				}

				savedmemory = null;
				setState(State.RUNNING);
				while (suspendedTasks.size() != 0) {
					suspendedTasks.remove(0).registerConsumption();
				}
			}
		}
		boolean trsuccess;
		if (bgnwload > 0) {
			startBGLoad();
			trsuccess = vasource.requestContentDelivery(savedmemory.id, tmemid,
					pmdisk, new ResumeComplete());
		} else {
			trsuccess = pmdisk.duplicateContent(savedmemory.id, tmemid,
					new ResumeComplete());
		}
		if (!trsuccess) {
			setState(State.SUSPENDED);
			throw new VMManagementException(
					"Not enough space on localDisk for the resume operation of "
							+ hashCode());
		}
	}

	public void subscribeStateChange(final StateChange consumer) {
		subscribers.add(consumer);
	}

	public void unsubscribeStateChange(final StateChange consumer) {
		subscribers.remove(consumer);
	}

	private class BGLoadEvent extends ConsumptionEventAdapter {
		@Override
		public void conComplete() {
			try {
				startBGLoad();
			} catch (NetworkException e) {
			}
		}
	}

	private BGLoadEvent loadEvent = new BGLoadEvent();

	private void startBGLoad() throws NetworkNode.NetworkException {
		if (consumingStates.contains(currState)) {
			final long bgnwload = va.getBgNetworkLoad();
			if (bgnwload > 0) {
				NetworkNode.initTransfer(bgnwload, host.getLocalDisk(),
						vasource, loadEvent);
			}
		}
	}

	public void changeResourceConstraints(final ResourceConstraints newRc) {
		rc = newRc;
		setPerSecondProcessingPower(newRc.requiredCPUs);
	}

	public void newComputeTask(final double total,
			final ResourceConsumption.ConsumptionEvent e)
			throws StateChangeException {
		if (!currState.equals(State.RUNNING)) {
			throw new StateChangeException(
					"Cannot process computing tasks while not in running state");
		}
		new ResourceConsumption(total, this, host, e).registerConsumption();
	}
}
